import { reactive } from 'vue';
import { cleanupForm, getResultFormData, path as _path, PromiseCancel, resetFields, set, transformFormData } from '../common';
import { Form } from '../form/Form';
import { ValidationError } from '../form/ValidationError';
/**
 *
 * @param formData The structure of your Form Data.
 * @description
 * Vue composition function for Form Validation.
 * @docs
 * https://github.com/JensDll/vue3-form-validation
 * @typescript
 * For better type inference, consider defining the structure
 * of your `formData` upfront and pass it as the generic parameter `T`:
 * ```
 * type FormData = {
 *   name: Field<string>,
 *   email: Field<string>,
 *   password: Field<string>
 * }
 *
 * const { ... } = useValidation<FormData>({ ... })
 * ```
 */
export function useValidation(formData) {
    const form = new Form();
    const promiseCancel = new PromiseCancel();
    transformFormData(form, formData);
    const transformedFormData = reactive(formData);
    return {
        form: transformedFormData,
        submitting: form.submitting,
        errors: form.errors,
        async validateFields(names) {
            form.submitting.value = true;
            const resultFormData = getResultFormData(transformedFormData);
            try {
                await promiseCancel.race(form.validateAll(names));
            }
            finally {
                form.submitting.value = false;
            }
            return resultFormData;
        },
        resetFields(formData) {
            if (form.submitting.value) {
                promiseCancel.cancelReject(new ValidationError());
            }
            if (formData) {
                form.resetFields(false);
                resetFields(formData, transformedFormData);
            }
            else {
                form.resetFields();
            }
        },
        add(path, value) {
            const lastKey = path[path.length - 1];
            if (typeof lastKey !== 'undefined') {
                const box = { [lastKey]: value };
                transformFormData(form, box);
                const x = _path(path, transformedFormData);
                if (Array.isArray(x)) {
                    x.push(box[lastKey]);
                }
                else {
                    set(transformedFormData, path, box[lastKey]);
                }
            }
        },
        remove(path) {
            const lastKey = path.pop();
            if (typeof lastKey !== 'undefined' && path.length === 0) {
                cleanupForm(form, transformedFormData[lastKey]);
                delete transformedFormData[lastKey];
            }
            else if (typeof lastKey !== 'undefined') {
                const value = _path(path, transformedFormData);
                if (Array.isArray(value)) {
                    const deleted = value.splice(+lastKey, 1);
                    cleanupForm(form, deleted);
                }
                else {
                    cleanupForm(form, value[lastKey]);
                    delete value[lastKey];
                }
            }
        }
    };
}
